@@ -46,6 +46,9 @@ True
 
 """
 
+from __future__ import absolute_import, division, print_function
+__metaclass__ = type
+
 import os
 import io
 import re
@@ -57,6 +60,11 @@ import traceback as traceback_
 import logging
 import platform
 
+try:
+    from functools import lru_cache
+except ImportError:
+    from backports.functools_lru_cache import lru_cache
+
 import six
 from six.moves import queue
 from six.moves import urllib
@@ -72,13 +80,31 @@ __all__ = ('HTTPRequest', 'HTTPConnection', 'HTTPServer',
            'Gateway', 'get_ssl_adapter_class')
 
 
-if 'win' in sys.platform and hasattr(socket, 'AF_INET6'):
+IS_WINDOWS = platform.system() == 'Windows'
+
+
+if not IS_WINDOWS:
+    import grp
+    import pwd
+    import struct
+
+
+if IS_WINDOWS and hasattr(socket, 'AF_INET6'):
     if not hasattr(socket, 'IPPROTO_IPV6'):
         socket.IPPROTO_IPV6 = 41
     if not hasattr(socket, 'IPV6_V6ONLY'):
         socket.IPV6_V6ONLY = 27
 
 
+if not hasattr(socket, 'SO_PEERCRED'):
+    """
+    NOTE: the value for SO_PEERCRED can be architecture specific, in
+    which case the getsockopt() will hopefully fail. The arch
+    specific value could be derived from platform.processor()
+    """
+    socket.SO_PEERCRED = 17
+
+
 LF = b'\n'
 CRLF = b'\r\n'
 TAB = b'\t'
@@ -106,7 +132,7 @@ if not hasattr(logging, 'statistics'):
     logging.statistics = {}
 
 
-class HeaderReader(object):
+class HeaderReader:
     """Object for reading headers from an HTTP request.
 
     Interface and default implementation.
@@ -180,7 +206,7 @@ class DropUnderscoreHeaderReader(HeaderReader):
         return orig and '_' not in key_name
 
 
-class SizeCheckWrapper(object):
+class SizeCheckWrapper:
     """Wraps a file-like object, raising MaxSizeExceeded if too large."""
 
     def __init__(self, rfile, maxlen):
@@ -206,6 +232,7 @@ class SizeCheckWrapper(object):
 
         Returns:
             bytes: Chunk from rfile, limited by size if specified.
+
         """
         data = self.rfile.read(size)
         self.bytes_read += len(data)
@@ -220,6 +247,7 @@ class SizeCheckWrapper(object):
 
         Returns:
             bytes: One line from rfile.
+
         """
         if size is not None:
             data = self.rfile.readline(size)
@@ -247,6 +275,7 @@ class SizeCheckWrapper(object):
 
         Returns:
             list[bytes]: Lines of bytes read from rfile.
+
         """
         # Shamelessly stolen from StringIO
         total = 0
@@ -275,15 +304,10 @@ class SizeCheckWrapper(object):
         self._check_length()
         return data
 
-    def next(self):
-        """Generate next file chunk."""
-        data = self.rfile.next()
-        self.bytes_read += len(data)
-        self._check_length()
-        return data
+    next = __next__
 
 
-class KnownLengthRFile(object):
+class KnownLengthRFile:
     """Wraps a file-like object, returning an empty string when exhausted."""
 
     def __init__(self, rfile, content_length):
@@ -292,6 +316,7 @@ class KnownLengthRFile(object):
         Args:
             rfile (file): file of a known size
             content_length (int): length of the file being read
+
         """
         self.rfile = rfile
         self.remaining = content_length
@@ -304,6 +329,7 @@ class KnownLengthRFile(object):
 
         Returns:
             bytes: Chunk from rfile, limited by size if specified.
+
         """
         if self.remaining == 0:
             return b''
@@ -324,6 +350,7 @@ class KnownLengthRFile(object):
 
         Returns:
             bytes: One line from rfile.
+
         """
         if self.remaining == 0:
             return b''
@@ -344,6 +371,7 @@ class KnownLengthRFile(object):
 
         Returns:
             list[bytes]: Lines of bytes read from rfile.
+
         """
         # Shamelessly stolen from StringIO
         total = 0
@@ -371,8 +399,10 @@ class KnownLengthRFile(object):
         self.remaining -= len(data)
         return data
 
+    next = __next__
 
-class ChunkedRFile(object):
+
+class ChunkedRFile:
     """Wraps a file-like object, returning an empty string when exhausted.
 
     This class is intended to provide a conforming wsgi.input value for
@@ -441,6 +471,7 @@ class ChunkedRFile(object):
 
         Returns:
             bytes: Chunk from rfile, limited by size if specified.
+
         """
         data = EMPTY
 
@@ -473,6 +504,7 @@ class ChunkedRFile(object):
 
         Returns:
             bytes: One line from rfile.
+
         """
         data = EMPTY
 
@@ -515,6 +547,7 @@ class ChunkedRFile(object):
 
         Returns:
             list[bytes]: Lines of bytes read from rfile.
+
         """
         # Shamelessly stolen from StringIO
         total = 0
@@ -533,6 +566,7 @@ class ChunkedRFile(object):
 
         Returns:
             Generator: yields CRLF separated lines.
+
         """
         if not self.closed:
             raise ValueError(
@@ -561,7 +595,7 @@ class ChunkedRFile(object):
         self.rfile.close()
 
 
-class HTTPRequest(object):
+class HTTPRequest:
     """An HTTP Request (and response).
 
     A single HTTP connection may consist of multiple request/response pairs.
@@ -667,6 +701,7 @@ class HTTPRequest(object):
 
         Returns:
             bool: True if the request line is valid or False if it's malformed.
+
         """
         # HTTP/1.1 connections are persistent by default. If a client
         # requests a page, then idles (leaves the connection open),
@@ -824,11 +859,11 @@ class HTTPRequest(object):
                 return False
 
             # Unquote the path+params (e.g. "/this%20path" -> "/this path").
-            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec5.html#sec5.1.2
+            # https://www.w3.org/Protocols/rfc2616/rfc2616-sec5.html#sec5.1.2
             #
             # But note that "...a URI must be separated into its components
             # before the escaped characters within those components can be
-            # safely decoded." http://www.ietf.org/rfc/rfc2396.txt, sec 2.4.2
+            # safely decoded." https://www.ietf.org/rfc/rfc2396.txt, sec 2.4.2
             # Therefore, "/this%2Fpath" becomes "/this%2Fpath", not
             # "/this/path".
             try:
@@ -888,7 +923,16 @@ class HTTPRequest(object):
             return False
 
         mrbs = self.server.max_request_body_size
-        if mrbs and int(self.inheaders.get(b'Content-Length', 0)) > mrbs:
+
+        try:
+            cl = int(self.inheaders.get(b'Content-Length', 0))
+        except ValueError:
+            self.simple_response(
+                '400 Bad Request',
+                'Malformed Content-Length Header.')
+            return False
+
+        if mrbs and cl > mrbs:
             self.simple_response(
                 '413 Request Entity Too Large',
                 'The entity sent with the request exceeds the maximum '
@@ -1106,7 +1150,7 @@ class HTTPRequest(object):
         self.conn.wfile.write(EMPTY.join(buf))
 
 
-class HTTPConnection(object):
+class HTTPConnection:
     """An HTTP connection (active socket)."""
 
     remote_addr = None
@@ -1115,6 +1159,8 @@ class HTTPConnection(object):
     rbufsize = io.DEFAULT_BUFFER_SIZE
     wbufsize = io.DEFAULT_BUFFER_SIZE
     RequestHandlerClass = HTTPRequest
+    peercreds_enabled = False
+    peercreds_resolve_enabled = False
 
     def __init__(self, server, sock, makefile=MakeFile):
         """Initialize HTTPConnection instance.
@@ -1131,6 +1177,18 @@ class HTTPConnection(object):
         self.wfile = makefile(sock, 'wb', self.wbufsize)
         self.requests_seen = 0
 
+        self.peercreds_enabled = self.server.peercreds_enabled
+        self.peercreds_resolve_enabled = self.server.peercreds_resolve_enabled
+
+        # LRU cached methods:
+        # Ref: https://stackoverflow.com/a/14946506/595220
+        self.resolve_peer_creds = (
+            lru_cache(maxsize=1)(self.resolve_peer_creds)
+        )
+        self.get_peer_creds = (
+            lru_cache(maxsize=1)(self.get_peer_creds)
+        )
+
     def communicate(self):
         """Read each request and respond appropriately."""
         request_seen = False
@@ -1228,6 +1286,102 @@ class HTTPConnection(object):
             # Apache does, but not today.
             pass
 
+    def get_peer_creds(self):  # LRU cached on per-instance basis, see __init__
+        """Return the PID/UID/GID tuple of the peer socket for UNIX sockets.
+
+        This function uses SO_PEERCRED to query the UNIX PID, UID, GID
+        of the peer, which is only available if the bind address is
+        a UNIX domain socket.
+
+        Raises:
+            NotImplementedError: in case of unsupported socket type
+            RuntimeError: in case of SO_PEERCRED lookup unsupported or disabled
+
+        """
+        PEERCRED_STRUCT_DEF = '3i'
+
+        if IS_WINDOWS or self.socket.family != socket.AF_UNIX:
+            raise NotImplementedError(
+                'SO_PEERCRED is only supported in Linux kernel and WSL'
+            )
+        elif not self.peercreds_enabled:
+            raise RuntimeError(
+                'Peer creds lookup is disabled within this server'
+            )
+
+        try:
+            peer_creds = self.socket.getsockopt(
+                socket.SOL_SOCKET, socket.SO_PEERCRED,
+                struct.calcsize(PEERCRED_STRUCT_DEF)
+            )
+        except socket.error as socket_err:
+            """Non-Linux kernels don't support SO_PEERCRED.
+
+            Refs:
+            http://welz.org.za/notes/on-peer-cred.html
+            https://github.com/daveti/tcpSockHack
+            msdn.microsoft.com/en-us/commandline/wsl/release_notes#build-15025
+            """
+            six.raise_from(  # 3.6+: raise RuntimeError from socket_err
+                RuntimeError,
+                socket_err,
+            )
+        else:
+            pid, uid, gid = struct.unpack(PEERCRED_STRUCT_DEF, peer_creds)
+            return pid, uid, gid
+
+    @property
+    def peer_pid(self):
+        """Return the id of the connected peer process."""
+        pid, _, _ = self.get_peer_creds()
+        return pid
+
+    @property
+    def peer_uid(self):
+        """Return the user id of the connected peer process."""
+        _, uid, _ = self.get_peer_creds()
+        return uid
+
+    @property
+    def peer_gid(self):
+        """Return the group id of the connected peer process."""
+        _, _, gid = self.get_peer_creds()
+        return gid
+
+    def resolve_peer_creds(self):  # LRU cached on per-instance basis
+        """Return the username and group tuple of the peercreds if available.
+
+        Raises:
+            NotImplementedError: in case of unsupported OS
+            RuntimeError: in case of UID/GID lookup unsupported or disabled
+
+        """
+        if IS_WINDOWS:
+            raise NotImplementedError(
+                'UID/GID lookup can only be done under UNIX-like OS'
+            )
+        elif not self.peercreds_resolve_enabled:
+            raise RuntimeError(
+                'UID/GID lookup is disabled within this server'
+            )
+
+        user = pwd.getpwuid(self.peer_uid).pw_name  # [0]
+        group = grp.getgrgid(self.peer_gid).gr_name  # [0]
+
+        return user, group
+
+    @property
+    def peer_user(self):
+        """Return the username of the connected peer process."""
+        user, _ = self.resolve_peer_creds()
+        return user
+
+    @property
+    def peer_group(self):
+        """Return the group of the connected peer process."""
+        _, group = self.resolve_peer_creds()
+        return group
+
     def _close_kernel_socket(self):
         """Close kernel socket in outdated Python versions.
 
@@ -1258,7 +1412,10 @@ except ImportError:
         _SetHandleInformation.restype = ctypes.wintypes.BOOL
     except ImportError:
         def prevent_socket_inheritance(sock):
-            """Dummy function, since neither fcntl nor ctypes are available."""
+            """Stub inheritance prevention.
+
+            Dummy function, since neither fcntl nor ctypes are available.
+            """
             pass
     else:
         def prevent_socket_inheritance(sock):
@@ -1273,7 +1430,7 @@ else:
         fcntl.fcntl(fd, fcntl.F_SETFD, old_flags | fcntl.FD_CLOEXEC)
 
 
-class HTTPServer(object):
+class HTTPServer:
     """An HTTP server."""
 
     _bind_addr = '127.0.0.1'
@@ -1342,9 +1499,17 @@ class HTTPServer(object):
     You must have the corresponding SSL driver library installed.
     """
 
+    peercreds_enabled = False
+    """If True, peer cred lookup can be performed via UNIX domain socket."""
+
+    peercreds_resolve_enabled = False
+    """If True, username/group will be looked up in the OS from peercreds."""
+
     def __init__(
-            self, bind_addr, gateway, minthreads=10, maxthreads=-1,
-            server_name=None):
+        self, bind_addr, gateway,
+        minthreads=10, maxthreads=-1, server_name=None,
+        peercreds_enabled=False, peercreds_resolve_enabled=False,
+    ):
         """Initialize HTTPServer instance.
 
         Args:
@@ -1364,6 +1529,10 @@ class HTTPServer(object):
         if not server_name:
             server_name = self.version
         self.server_name = server_name
+        self.peercreds_enabled = peercreds_enabled
+        self.peercreds_resolve_enabled = (
+            peercreds_resolve_enabled and peercreds_enabled
+        )
         self.clear_stats()
 
     def clear_stats(self):
@@ -1576,7 +1745,7 @@ class HTTPServer(object):
         """Create (or recreate) the actual socket object."""
         self.socket = socket.socket(family, type, proto)
         prevent_socket_inheritance(self.socket)
-        if platform.system() != 'Windows':
+        if not IS_WINDOWS:
             # Windows has different semantics for SO_REUSEADDR,
             # so don't set it.
             # https://msdn.microsoft.com/en-us/library/ms740621(v=vs.85).aspx
@@ -1607,7 +1776,19 @@ class HTTPServer(object):
                 pass
 
         self.socket.bind(self.bind_addr)
-        self.bind_addr = self.socket.getsockname()[:2]  # TODO: keep separate
+        # TODO: keep requested bind_addr separate real bound_addr (port is
+        # different in case of ephemeral port 0)
+        self.bind_addr = self.socket.getsockname()
+        if family in (
+            # Windows doesn't have socket.AF_UNIX, so not using it in check
+            socket.AF_INET,
+            socket.AF_INET6,
+        ):
+            """UNIX domain sockets are strings or bytes.
+
+            In case of bytes with a leading null-byte it's an abstract socket.
+            """
+            self.bind_addr = self.bind_addr[:2]
 
     def tick(self):
         """Accept a new connection and put it on the Queue."""
@@ -1744,7 +1925,7 @@ class HTTPServer(object):
                         try:
                             s = socket.socket(af, socktype, proto)
                             # See
-                            # http://groups.google.com/group/cherrypy-users/
+                            # https://groups.google.com/group/cherrypy-users/
                             #     browse_frm/thread/bbfe5eb39c904fe0
                             s.settimeout(1.0)
                             s.connect((host, port))
@@ -1759,7 +1940,7 @@ class HTTPServer(object):
         self.requests.stop(self.shutdown_timeout)
 
 
-class Gateway(object):
+class Gateway:
     """Base class to interface HTTPServer with other systems, such as WSGI."""
 
     def __init__(self, req):
